package org.apache.bcel.verifier.structurals;

package .tmp;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.bcel.generic.BranchInstruction;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.Select;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;
import org.apache.bcel.verifier.structurals.ControlFlowGraph;
import org.apache.bcel.verifier.structurals.ExceptionHandler;
import org.apache.bcel.verifier.structurals.ExecutionVisitor;
import org.apache.bcel.verifier.structurals.Frame;
import org.apache.bcel.verifier.structurals.InstConstraintVisitor;
import org.apache.bcel.verifier.structurals.InstructionContext;
import org.apache.bcel.verifier.structurals.LocalVariables;
import org.apache.bcel.verifier.structurals.OperandStack;
import org.apache.bcel.verifier.structurals.Subroutine;

class InstructionContextImpl implements InstructionContext {
  private int TAG;
  
  private final InstructionHandle instruction;
  
  private final Map<InstructionContext, Frame> inFrames;
  
  private final Map<InstructionContext, Frame> outFrames;
  
  private List<InstructionContext> executionPredecessors;
  
  public InstructionContextImpl(InstructionHandle inst) {
    if (inst == null)
      throw new AssertionViolatedException("Cannot instantiate InstructionContextImpl from NULL."); 
    this.instruction = inst;
    this.inFrames = new HashMap<>();
    this.outFrames = new HashMap<>();
  }
  
  private InstructionHandle[] _getSuccessors() {
    InstructionHandle[] single = new InstructionHandle[1];
    Instruction inst = getInstruction().getInstruction();
    if (inst instanceof org.apache.bcel.generic.RET) {
      Subroutine s = ControlFlowGraph.access$000(ControlFlowGraph.this).subroutineOf(getInstruction());
      if (s == null)
        throw new AssertionViolatedException("Asking for successors of a RET in dead code?!"); 
      InstructionHandle[] jsrs = s.getEnteringJsrInstructions();
      InstructionHandle[] ret = new InstructionHandle[jsrs.length];
      Arrays.setAll(ret, i -> jsrs[i].getNext());
      return ret;
    } 
    if (inst instanceof org.apache.bcel.generic.ReturnInstruction || inst instanceof org.apache.bcel.generic.ATHROW)
      return InstructionHandle.EMPTY_ARRAY; 
    if (inst instanceof JsrInstruction) {
      single[0] = ((JsrInstruction)inst).getTarget();
      return single;
    } 
    if (inst instanceof GotoInstruction) {
      single[0] = ((GotoInstruction)inst).getTarget();
      return single;
    } 
    if (inst instanceof BranchInstruction) {
      if (inst instanceof Select) {
        InstructionHandle[] matchTargets = ((Select)inst).getTargets();
        InstructionHandle[] ret = new InstructionHandle[matchTargets.length + 1];
        ret[0] = ((Select)inst).getTarget();
        System.arraycopy(matchTargets, 0, ret, 1, matchTargets.length);
        return ret;
      } 
      InstructionHandle[] pair = new InstructionHandle[2];
      pair[0] = getInstruction().getNext();
      pair[1] = ((BranchInstruction)inst).getTarget();
      return pair;
    } 
    single[0] = getInstruction().getNext();
    return single;
  }
  
  public boolean execute(Frame inFrame, ArrayList<InstructionContext> execPreds, InstConstraintVisitor icv, ExecutionVisitor ev) {
    List<InstructionContext> clone = (List<InstructionContext>)execPreds.clone();
    this.executionPredecessors = clone;
    if ((lastExecutionJSR() == null && ControlFlowGraph.access$000(ControlFlowGraph.this).subroutineOf(getInstruction()) != ControlFlowGraph.access$000(ControlFlowGraph.this).getTopLevel()) || (
      lastExecutionJSR() != null && ControlFlowGraph.access$000(ControlFlowGraph.this).subroutineOf(getInstruction()) == ControlFlowGraph.access$000(ControlFlowGraph.this).getTopLevel()))
      throw new AssertionViolatedException("Huh?! Am I '" + this + "' part of a subroutine or not?"); 
    Frame inF = this.inFrames.get(lastExecutionJSR());
    if (inF == null) {
      this.inFrames.put(lastExecutionJSR(), inFrame);
      inF = inFrame;
    } else if (inF.equals(inFrame) || !mergeInFrames(inFrame)) {
      return false;
    } 
    Frame workingFrame = inF.getClone();
    try {
      icv.setFrame(workingFrame);
      getInstruction().accept(icv);
    } catch (StructuralCodeConstraintException ce) {
      ce.extendMessage("", "\nInstructionHandle: " + getInstruction() + "\n");
      ce.extendMessage("", "\nExecution Frame:\n" + workingFrame);
      extendMessageWithFlow(ce);
      throw ce;
    } 
    ev.setFrame(workingFrame);
    getInstruction().accept(ev);
    this.outFrames.put(lastExecutionJSR(), workingFrame);
    return true;
  }
  
  private void extendMessageWithFlow(StructuralCodeConstraintException e) {
    String s = "Execution flow:\n";
    e.extendMessage("", "Execution flow:\n" + getExecutionChain());
  }
  
  public ExceptionHandler[] getExceptionHandlers() {
    return ControlFlowGraph.access$100(ControlFlowGraph.this).getExceptionHandlers(getInstruction());
  }
  
  private String getExecutionChain() {
    StringBuilder s = new StringBuilder(toString());
    for (int i = this.executionPredecessors.size() - 1; i >= 0; i--)
      s.insert(0, (new StringBuilder()).append(this.executionPredecessors.get(i)).append("\n").toString()); 
    return s.toString();
  }
  
  public Frame getInFrame() {
    InstructionContext jsr = lastExecutionJSR();
    Frame org = this.inFrames.get(jsr);
    if (org == null)
      throw new AssertionViolatedException("inFrame not set! This:\n" + this + "\nInFrames: '" + this.inFrames + "'."); 
    return org.getClone();
  }
  
  public InstructionHandle getInstruction() {
    return this.instruction;
  }
  
  public Frame getOutFrame(ArrayList<InstructionContext> execChain) {
    this.executionPredecessors = execChain;
    InstructionContext jsr = lastExecutionJSR();
    Frame org = this.outFrames.get(jsr);
    if (org == null)
      throw new AssertionViolatedException("outFrame not set! This:\n" + this + "\nExecutionChain: " + 
          getExecutionChain() + "\nOutFrames: '" + this.outFrames + "'."); 
    return org.getClone();
  }
  
  public InstructionContext[] getSuccessors() {
    return ControlFlowGraph.this.contextsOf(_getSuccessors());
  }
  
  public int getTag() {
    return this.TAG;
  }
  
  private InstructionContextImpl lastExecutionJSR() {
    int size = this.executionPredecessors.size();
    int retcount = 0;
    for (int i = size - 1; i >= 0; i--) {
      InstructionContextImpl current = (InstructionContextImpl)this.executionPredecessors.get(i);
      Instruction currentlast = current.getInstruction().getInstruction();
      if (currentlast instanceof org.apache.bcel.generic.RET)
        retcount++; 
      retcount--;
      if (currentlast instanceof JsrInstruction && retcount == -1)
        return current; 
    } 
    return null;
  }
  
  private boolean mergeInFrames(Frame inFrame) {
    Frame inF = this.inFrames.get(lastExecutionJSR());
    OperandStack oldstack = inF.getStack().getClone();
    LocalVariables oldlocals = inF.getLocals().getClone();
    try {
      inF.getStack().merge(inFrame.getStack());
      inF.getLocals().merge(inFrame.getLocals());
    } catch (StructuralCodeConstraintException sce) {
      extendMessageWithFlow(sce);
      throw sce;
    } 
    return (!oldstack.equals(inF.getStack()) || !oldlocals.equals(inF.getLocals()));
  }
  
  public void setTag(int tag) {
    this.TAG = tag;
  }
  
  public String toString() {
    return getInstruction().toString(false) + "\t[InstructionContext]";
  }
}
